
#include "./XvidEncoder.h"
#include "./Settings.h"

#include <stdio.h>
#include <shlobj.h>


XvidEncoder::XvidEncoder(HINSTANCE dllInstance, HWND owner) : 
    _dllInstance(dllInstance),
    _owner(owner),
    _outputFile(0),
    _audioFrameAccumulation(0)
{
    setDefaultOptions();
    loadOptions();
}



XvidEncoder::~XvidEncoder() {
    saveOptions();
    clearOptions();

    delete _outputFile;
    delete _audioFrameAccumulation;    
}



void XvidEncoder::showOptions() {
    XvidSettings::Show(_dllInstance, _owner, this);
}



bool XvidEncoder::onStartEncoding(const char* movieFile) {
    // Get output filename
    {
        char* movieDirectory = new char[strlen(movieFile) + 1]; strcpy(movieDirectory, movieFile); //strdup(movieFile);
        int i = (int)strlen(movieDirectory) - 1;
        for (; i >= 0; i--) { if (movieDirectory[i] == '\\' || movieDirectory[i] == '/') break; }
        if (i >= 0) movieDirectory[i] = '\0';

        char output[4096] = "";
        OPENFILENAME ofn;
	    ZeroMemory(&ofn, sizeof(ofn));
	    ofn.lStructSize = sizeof(ofn);
	    ofn.hwndOwner = 0; // TODO
	    ofn.lpstrFilter = ".AVI Files\0*.avi\0All Files\0*.*\0";
	    ofn.lpstrFile = output;
	    ofn.nMaxFile = 4096;
	    ofn.Flags = OFN_EXPLORER | OFN_HIDEREADONLY;
	    ofn.lpstrDefExt = "avi";									
        ofn.lpstrTitle = "Save encoded movie as...";

        
        for (unsigned int i = 0; i < strlen(movieDirectory); i++) {
            if (movieDirectory[i] == '/') movieDirectory[i] = '\\';
        }
        ofn.lpstrInitialDir = movieDirectory;

        if (GetSaveFileName(&ofn) != TRUE) return false;

        delete movieDirectory;

        _outputFile = new char[strlen(output) + 1]; strcpy(_outputFile, output); //strdup(output);
    }

    float videoQuality;
    sscanf(getOption(OPTION_VIDEO_QUALITY), "%f", &videoQuality);

    int videoFramerate;
    sscanf(getOption(OPTION_VIDEO_FRAMERATE), "%i", &videoFramerate);
    switch (videoFramerate) {
        case 15: _videoFrameDivider = 4; break;
        case 20: _videoFrameDivider = 3; break;
        case 30: _videoFrameDivider = 2; break;
        case 60:
        default: _videoFrameDivider = 1; break;
    }
    _videoFrameCounter = 1;


    int audioSamplerate;
    sscanf(getOption(OPTION_AUDIO_SAMPLERATE), "%i", &audioSamplerate);
    switch (audioSamplerate) {
        case 11025: _audioFrameDivider = 4; break;
        case 22050: _audioFrameDivider = 2; break;
        case 44100:
        default: _audioFrameDivider = 1; break;
    }
    _audioFrameCounter = _audioFrameDivider;
    audioSamplerate = 44640 / _audioFrameDivider; // correct for slight quirk in nes sound output (base freq is not exactly 44 khz)

    // If downsampling, we need to accumulate audio frames and then periodically downsample them and send them to the encoder
    int audioFrameSize = 1488;
    _audioFrameAccumulation = new unsigned char[audioFrameSize * (_audioFrameDivider + 2)];
    _audioAccumulationOffset = 0;

    // Initialize Revel
    Revel_Error revError = Revel_CreateEncoder(&_encoderHandle);

    Revel_Params revParams;
    Revel_InitializeParams(&revParams);
    revParams.width = 256;
    revParams.height = 224;
    revParams.frameRate = (float)videoFramerate;
    revParams.quality = videoQuality;
    revParams.codec = REVEL_CD_XVID;

    revParams.hasAudio = 1;
    revParams.audioChannels = 1;
    revParams.audioRate = audioSamplerate;
    revParams.audioBits = 16;
    revParams.audioSampleFormat = REVEL_ASF_PCM;

    // Initiate encoding
    revError = Revel_EncodeStart(_encoderHandle, _outputFile, &revParams);

    return true;
}



void XvidEncoder::onFrame(const unsigned char* videoFrame, const unsigned char* audioFrame) {
    Revel_Error revError;

    // Encode video
    {
        _videoFrameCounter--;
        if (!_videoFrameCounter) {
            _videoFrameCounter = _videoFrameDivider;
            
            Revel_VideoFrame frame;
            frame.width = 256;
            frame.height = 224;
            frame.bytesPerPixel = 3;
            frame.pixelFormat = REVEL_PF_BGR;
            frame.pixels = (void*)videoFrame;

            int frameSize;
            Revel_Error revError = Revel_EncodeFrame(_encoderHandle, &frame, &frameSize);
            if (revError != REVEL_ERR_NONE) {
                // might want to handle error
            }
        }
    }

    // Encode audio
    {
        // Accumulate audio frame into 44 KHz buffer
        memcpy(&_audioFrameAccumulation[_audioAccumulationOffset], (void*)audioFrame, 1488);
        _audioAccumulationOffset += 1488;

        _audioFrameCounter--;
        if (!_audioFrameCounter) {
            _audioFrameCounter = _audioFrameDivider;

            // Filter accumulated sound down to proper size
            downsampleAudio(_audioFrameAccumulation, _audioAccumulationOffset/2, (unsigned char*)audioFrame, 1488/2);
            _audioAccumulationOffset = 0;

            int totalAudioBytes = 0;
            revError = Revel_EncodeAudio(_encoderHandle, (void*)audioFrame, 1488, &totalAudioBytes);
        }
    }

}



bool XvidEncoder::onStopEncoding() {
    int totalSize;
    Revel_Error revError = Revel_EncodeEnd(_encoderHandle, &totalSize);
    if (revError != REVEL_ERR_NONE) {
        
    }

    Revel_DestroyEncoder(_encoderHandle);
 
    return true;
}



// bad downsammpling implementation
void XvidEncoder::downsampleAudio(unsigned char* input, int inputNumberSamples, unsigned char* output, int outputNumberSamples) {
    
    int downsampleFactor = inputNumberSamples / outputNumberSamples;

    unsigned short* inSample = (unsigned short*)input;
    unsigned short* outSample = (unsigned short*)output;
    for (int i = 0; i < outputNumberSamples; i++) {
        float accum = 0;
        for (int j = 0; j < downsampleFactor; j++) {
            accum += *inSample++;
        }
        accum /= downsampleFactor;
        *outSample++ = (unsigned short)accum;
    }
}



char* XvidEncoder::getOption(char* key) {
    std::map<char*, char*, ltstr>::iterator mit = _settings.find(key);
    return (mit != _settings.end()) ? mit->second : 0;
}



void XvidEncoder::setOption(char* key, char* value) {    
    std::map<char*, char*, ltstr>::iterator mit = _settings.find(key);
    char* localValue = new char[strlen(value) + 1];
    strcpy(localValue, value);
    if (mit != _settings.end()) {
        delete mit->second;
        mit->second = localValue;
    } else {
        char* localKey = new char[strlen(key) + 1];
        strcpy(localKey, key);
        _settings[localKey] = localValue;
    }
}



void XvidEncoder::loadOptions() {
    char dllPath[4096];
	GetModuleFileName(_dllInstance, dllPath, 4096);
    strcat(dllPath, ".cfg");
    FILE* s = fopen(dllPath, "r");
    if (!s) return;

    while (!feof(s)) {
        char line[2048];
        fgets(line, 2048, s);
        while (line[strlen(line)-1]=='\n') line[strlen(line)-1] = '\0';
        char* sep = strstr(line, "=");
        if (!sep) continue;
        *sep++ = '\0';
        setOption(line, sep);
    }

    fclose(s);
}



void XvidEncoder::saveOptions() {
    char dllPath[4096];
	GetModuleFileName(_dllInstance, dllPath, 4096);
    strcat(dllPath, ".cfg");

    FILE* s = fopen(dllPath, "w");
    if (!s) return;

    for (std::map<char*, char*, ltstr>::iterator mit = _settings.begin(); mit != _settings.end(); mit++) {
        fprintf(s, "%s=%s\n", mit->first, mit->second);
    }

    fclose(s);
}



void XvidEncoder::clearOptions() {
    for (std::map<char*, char*, ltstr>::iterator mit = _settings.begin(); mit != _settings.end(); mit++) {
        delete mit->first;
        delete mit->second;
    }
    _settings.clear();
}



void XvidEncoder::setDefaultOptions() {
    setOption(OPTION_VIDEO_QUALITY, "1.0");
    setOption(OPTION_VIDEO_FRAMERATE, "60");
    setOption(OPTION_AUDIO_SAMPLERATE, "44100");
}